# Copyright (c) 2019 Capital Fund Management
# SPDX-License-Identifier: MIT

import argparse
import hashlib
import logging
import os.path
import socket
from configparser import ConfigParser, NoSectionError, NoOptionError

from tornado.ioloop import IOLoop
from tornado.web import StaticFileHandler, Application

from jupytab_server.jupytab_api import RestartHandler, APIHandler, EvaluateHandler, \
    ReverseProxyHandler, root, api_kernel, access_kernel, restart_kernel
from jupytab_server.kernel_executor import KernelExecutor
from jupytab_server.structures import CaseInsensitiveDict

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)


def __extract_item(config, notebook_key, item_key, default=None):
    try:
        return config.get(notebook_key, item_key)
    except (NoOptionError, NoSectionError):
        if default:
            return default
        else:
            raise ValueError(f'Expecting {item_key} in section {notebook_key}')


def config_listen_port(config):
    return config.getint('main', 'listen_port')


def config_security_token(config):
    try:
        return config.get('main', 'security_token')
    except (NoSectionError, NoOptionError):
        return None


def config_notebooks(config):
    notebooks = config.get('main', 'notebooks')
    notebook_dict = {}

    for key in notebooks.split('|'):
        nb_path = __extract_item(config, key, 'path')
        nb_name = __extract_item(config, key, 'name', key)
        nb_description = __extract_item(config, key, 'description', "No description")
        nb_cwd = __extract_item(config, key, 'directory', ".")

        notebook_dict[key] = {'name': nb_name,
                              'file_path': nb_path,
                              'description': nb_description,
                              'cwd': nb_cwd}

    return notebook_dict


def config_ssl(config):
    try:
        ssl_enabled = config.getboolean('main', 'ssl_enabled')
    except (NoSectionError, NoOptionError):
        ssl_enabled = False

    if ssl_enabled:
        ssl_cert = config.get('main', 'ssl_cert')
        ssl_key = config.get('main', 'ssl_key')

        if ssl_enabled and not os.path.isfile(ssl_cert):
            raise FileNotFoundError(f"SSL enabled but missing ssl_cert file: {ssl_cert}")
        if ssl_enabled and not os.path.isfile(ssl_key):
            raise FileNotFoundError(f"SSL enabled but missing ssl_key file: {ssl_key}")

        return {
            "certfile": ssl_cert,
            "keyfile": ssl_key
        }
    else:
        return None


def parse_config(config_file):
    if not os.path.isfile(config_file):
        raise FileNotFoundError(f"missing configuration: {config_file}")

    config = ConfigParser()
    config.optionxform = str
    config.read(config_file)

    listen_port = config_listen_port(config)
    security_token = config_security_token(config)
    notebooks = config_notebooks(config)
    ssl = config_ssl(config)

    if not ssl:
        print("SSL not enabled")
    else:
        print("SSL enabled")

    return {
        'listen_port': listen_port,
        'security_token': security_token,
        'notebooks': notebooks,
        'ssl': ssl
    }


def create_server_app(listen_port, security_token, notebooks, ssl):
    notebook_store = CaseInsensitiveDict()

    for key, value in notebooks.items():
        notebook_store[key] = KernelExecutor(**value)

    for key, value in notebook_store.items():
        value.start()

    protocol = "https" if ssl else "http"

    if security_token:
        token_digest = hashlib.sha224(security_token.encode('utf-8')).hexdigest()
        print(f"""Your token is {token_digest}
        Please open : {protocol}://{socket.gethostname()}:{listen_port}"""
              f"/?security_token={token_digest}")
    else:
        token_digest = None
        print(f"""You have no defined token. Please note your process is not secured !
        Please open : {protocol}://{socket.gethostname()}:{listen_port}""")

    server_app = Application([
        (r"/evaluate", EvaluateHandler,
         {'notebook_store': notebook_store, 'security_token': token_digest}),
        (r"/" + api_kernel, APIHandler,
         {'notebook_store': notebook_store, 'security_token': token_digest}),
        (r"/" + restart_kernel + "/(.*)", RestartHandler,
         {'notebook_store': notebook_store, 'security_token': token_digest}),
        (r"/" + access_kernel + "/(.*)", ReverseProxyHandler,
         {'notebook_store': notebook_store, 'security_token': token_digest}),

        (r"/(.*)", StaticFileHandler, {'path': root, "default_filename": "index.html"}),

    ])
    return server_app


def main(input_args=None):
    parser = argparse.ArgumentParser(description='The Tableau gateway to notebooks')
    parser.add_argument("-c", "--config", dest='config_file', default='config.ini', type=str)
    parser.add_argument("-e", "--env", dest='environment', default='UNKNOWN', type=str)
    args, unknown = parser.parse_known_args(args=input_args)

    params = parse_config(config_file=args.config_file)

    app = create_server_app(**params)

    app.listen(params['listen_port'], ssl_options=params['ssl'])

    IOLoop.instance().start()


if __name__ == "__main__":
    main()



import os
from configparser import ConfigParser

import pytest

from jupytab_server.jupytab import config_ssl

dir_path = os.path.dirname(os.path.realpath(__file__))
os.chdir(dir_path)


def config(config_file):
    config = ConfigParser()
    config.optionxform = str
    config.read(config_file)
    return config


def test_ssl():
    with pytest.raises(FileNotFoundError):
        config_ssl(config('config/ssl_bad_file.ini'))
    assert config_ssl(config('config/ssl_disabled.ini')) is None
    assert config_ssl(config('config/ssl_none.ini')) is None
    assert config_ssl(config('config/ssl_ok.ini')) == {'certfile': 'config/example.cert',
                                                       'keyfile': 'config/example.key'}



# Copyright (c) 2019 Capital Fund Management
# SPDX-License-Identifier: MIT

import logging
import os
import threading


class LogPipe(threading.Thread):

    def __init__(self, level):
        """Setup the object with a logger and a loglevel
        and start the thread
        """
        threading.Thread.__init__(self)
        self.daemon = True
        self.level = level
        self.fdRead, self.fdWrite = os.pipe()
        self.pipeReader = os.fdopen(self.fdRead)
        self.start()

    def fileno(self):
        """Return the write file descriptor of the pipe
        """
        return self.fdWrite

    def run(self):
        """Run the thread, logging everything.
        """
        for line in iter(self.pipeReader.readline, ''):
            logging.log(self.level, line.strip('\n'))

        self.pipeReader.close()

    def close(self):
        """Close the write end of the pipe.
        """
        os.close(self.fdWrite)



# Copyright (c) 2019 Capital Fund Management
# SPDX-License-Identifier: MIT

import json
import os
import traceback
from typing import Any
from urllib.parse import urlunparse, urlencode

from tornado.httpclient import AsyncHTTPClient, HTTPRequest
from tornado.web import RequestHandler, HTTPError

root = os.path.dirname(__file__) + '/static'
api_kernel = 'api'
access_kernel = 'kernel'
restart_kernel = 'api/restart'
uri_security_token = 'security_token'
evaluate_method = "evaluate"


def transform_url(protocol, host, path, query=''):
    new_path = urlunparse((protocol, host, path, '', query, ''))
    return new_path


class BaseRequestHandler(RequestHandler):

    def initialize(self, notebook_store, security_token):
        self.notebook_store = notebook_store
        self.security_token = security_token

    def set_default_headers(self) -> None:
        self.set_header(name="Content-Type", value="application/json; charset=UTF-8")

    def write_error(self, status_code: int, **kwargs: Any) -> None:
        body = {
            'method': self.request.method,
            'uri': self.request.path,
            'code': status_code,
            'message': self._reason
        }
        if "exc_info" in kwargs:
            if self.settings.get("serve_traceback"):
                # in debug mode, send a traceback
                trace = '\n'.join(traceback.format_exception(
                    *kwargs['exc_info']
                ))
                body['trace'] = trace
            if isinstance(kwargs['exc_info'][1], HTTPError):
                body['message'] = kwargs['exc_info'][1].log_message
        self.finish(body)

    def on_chunk(self, chunk):
        self.write(chunk)
        self.flush()


class EvaluateHandler(BaseRequestHandler):

    async def post(self, *args, **kwargs):
        query = json.loads(self.request.body)

        # api_key = query['api_key']
        script = query['script']
        data = query['data']

        if script == 'return _arg1' and data['_arg1'] == 1:
            # Query sent by Tableau to test connection ...
            # b'{"api_key":"","script":"return _arg1","data":{"_arg1":1}}'
            self.write(json.dumps(1))
        else:
            script_params = query['script'].split('.')
            if len(script_params) != 2:
                raise HTTPError(log_message=f"<Notebook.function> expected instead of <{script}>")

            notebook_target, method_target = script_params

            if notebook_target not in self.notebook_store:
                raise HTTPError(log_message=f"Unknown notebook ({notebook_target})")

            kernel = self.notebook_store[notebook_target]
            host = f'{kernel.host}:{kernel.port}'
            my_url = transform_url('http', host, evaluate_method)

            body = {
                'function': method_target.lower(),
                'data': data
            }
            json_body = json.dumps(body)

            request = HTTPRequest(
                url=my_url,
                method='POST',
                headers=self.request.headers,
                follow_redirects=True,
                body=json_body,
                request_timeout=3600.0,
                streaming_callback=self.on_chunk)

            response = await AsyncHTTPClient().fetch(request)

            self.set_status(response.code)
            self.finish()


class RestartHandler(BaseRequestHandler):
    def get(self, *args, **kwargs):
        if self.security_token and self.get_argument(uri_security_token) != self.security_token:
            raise ConnectionRefusedError("Invalid security token")

        kernel_id = self.request.path[len(restart_kernel) + 2:]

        notebook_kernel = self.notebook_store[kernel_id]

        notebook_kernel.restart()


class APIHandler(BaseRequestHandler):
    def get(self, *args, **kwargs):
        if self.security_token and self.get_argument(uri_security_token) != self.security_token:
            raise ConnectionRefusedError("Invalid security token")

        notebook_dict = {}

        for key, value in self.notebook_store.items():
            notebook_dict[key] = \
                {
                    "kernel_id": key,
                    "name": value.name,
                    "status": value.kernel_status,
                    "path": str(value.notebook_file),
                    "host": value.host,
                    "port": value.port,
                    "description": value.description
                }

        self.write(notebook_dict)


class ReverseProxyHandler(BaseRequestHandler):

    async def get(self, *args, **kwargs):
        if self.security_token and self.get_argument(uri_security_token) != self.security_token:
            raise ConnectionRefusedError("Invalid security token")

        notebook_path = self.request.path[len(access_kernel) + 2:].split('/', 2)

        notebook_id = notebook_path[0]
        notebook_uri = notebook_path[1] if len(notebook_path) > 1 else ''
        kernel = self.notebook_store[notebook_id]

        host = f'{kernel.host}:{kernel.port}'

        query_arguments = self.request.query_arguments.copy()
        query_arguments.pop(uri_security_token, None)

        my_url = transform_url(
            'http',
            host,
            notebook_uri,
            urlencode(query_arguments, doseq=True))

        request = HTTPRequest(
            url=my_url,
            method='GET',
            headers=self.request.headers,
            follow_redirects=True,
            body=None,
            request_timeout=3600.0,
            streaming_callback=self.on_chunk)

        response = await AsyncHTTPClient().fetch(request)

        self.set_status(response.code)
        self.finish()



import asyncio
import os
import threading
import time
from json.decoder import JSONDecodeError

import pytest
import requests
from tornado.ioloop import IOLoop

from jupytab_server.jupytab import parse_config, create_server_app

THIS_DIR = os.path.abspath(os.path.dirname(__file__))
RESOURCES = os.path.abspath(os.path.join(THIS_DIR, '..'))

PROTOCOL = "http"
HOST = "127.0.0.1"
PORT = "8765"
SECURITY_TOKEN = "02014868fe0eef123269397c5bc65a9608b3cedb73e3b84d8d02c220"

event_loop_instance = None


def run_jupytab():
    # Required as we run the server in a new thread
    asyncio.set_event_loop(asyncio.new_event_loop())
    os.chdir(os.path.abspath(RESOURCES))

    params = parse_config(config_file="samples/config.ini")

    server_app = create_server_app(**params)
    server_app.listen(params['listen_port'])

    global event_loop_instance
    event_loop_instance = IOLoop.instance()
    event_loop_instance.start()


@pytest.fixture(scope='session', autouse=True)
def setup_jupytab():
    # Setup
    thread = threading.Thread(target=run_jupytab)
    thread.daemon = True
    thread.start()

    yield

    # Teardown
    global event_loop_instance
    # Cleanly shutdown the event loop used for testing server
    event_loop_instance.add_callback(event_loop_instance.stop)


def build_uri(uri, params=None, protocol=PROTOCOL, host=HOST, port=PORT, token=SECURITY_TOKEN):
    token = f"security_token={token}" if token else ""
    return f"{protocol}://{host}:{port}/{uri}?{token}&{params}"


def rget(uri, delay_seconds=5, attempt_count=6):
    for attempt in range(attempt_count):
        try:
            response = requests.get(uri)
            if response.status_code == 200:
                return response.json()
        except requests.exceptions.RequestException as e:
            print("RequestException", e)
        except JSONDecodeError as e:
            print("JSONDecodeError", e)

        time.sleep(delay_seconds)

    raise ConnectionError(f"Unable to retrieve GET datas from {uri}")


def rpost(uri, body, delay_seconds=5, attempt_count=6):
    for attempt in range(attempt_count):
        try:
            response = requests.post(uri, json=body)
            if response.status_code == 200:
                return response.json()
        except requests.exceptions.RequestException as e:
            print("RequestException", e)
        except JSONDecodeError as e:
            print("JSONDecodeError", e)

        time.sleep(delay_seconds)

    raise ConnectionError(f"Unable to retrieve POST datas from {uri}")


def get_table_by_id(tables, id):
    for table in tables:
        if table['id'] == id:
            return table
    return None


def get_column_by_id(columns, id):
    for column in columns:
        if column['id'] == id:
            return column
    return None


def test_airflights_schema():
    response = rget(build_uri("kernel/AirFlights/schema"))

    json_result = response

    flight_table = get_table_by_id(json_result, 'flights')
    assert get_column_by_id(flight_table['columns'], "callsign") is not None
    velocity = get_column_by_id(flight_table['columns'], "velocity")
    assert velocity['dataType'] == "float"

    assert len(json_result[0]['columns']) == 16


def test_airflights_data():
    response = rget(build_uri("kernel/AirFlights/data", "table_name=flights"))

    json_result = response

    assert len(json_result) > 0
    assert 'callsign' in json_result[0]
    assert 'time_position' in json_result[0]
    assert 'baro_altitude' in json_result[0]


def test_realestatecrime_schema():
    response = rget(build_uri("kernel/RealEstateCrime/schema"))

    json_result = response

    crime_table = get_table_by_id(json_result, 'crime')
    assert len(crime_table['columns']) == 9
    assert get_column_by_id(crime_table['columns'], 'beat') is not None

    real_estate_table = get_table_by_id(json_result, 'real_estate')
    assert len(real_estate_table['columns']) == 12
    city = get_column_by_id(real_estate_table['columns'], 'city')
    assert city['dataType'] == 'string'

    combined = get_table_by_id(json_result, 'combined')
    assert len(combined['columns']) == 19
    baths = get_column_by_id(real_estate_table['columns'], 'baths')
    assert baths['dataType'] == 'int'


def test_realestatecrime_data():
    response = rget(build_uri("kernel/RealEstateCrime/data", "table_name=combined"))

    json_result = response

    assert len(json_result) > 0
    assert 'address' in json_result[0]
    assert 'beat' in json_result[0]
    assert 'city' in json_result[0]


def test_sklearn_classifier_schema():
    response = rget(build_uri("kernel/SKLearnClassifier/schema"))

    json_result = response

    assert len(json_result) == 3

    assert get_table_by_id(json_result, 'iris_target') is not None


def test_sklearn_classifier_data():
    response = rget(build_uri("kernel/SKLearnClassifier/data", "table_name=iris"))

    json_result = response

    assert len(json_result) > 100
    assert len(json_result[0]) == 5


def test_sklearn_classifier_function():
    predict_function = {
        "script": "SKLearnClassifier.predict",
        "data": {
            "_arg1": [5.1, 5.8, 6.5],
            "_arg2": [3.5, 2.6, 3],
            "_arg3": [1.4, 4, 5.5],
            "_arg4": [0.2, 1.2, 2.4],
        }
    }

    response = rpost(build_uri("evaluate"), predict_function)

    assert response == ["setosa", "versicolor", "virginica"]



# Copyright (c) Capital Fund Management
# Distributed under the terms of the MIT License.

import os
from setuptools import setup, find_packages

with open(os.path.join(os.path.dirname(__file__), '..', 'VERSION')) as version_file:
    VERSION = version_file.read().strip()

README = """

Jupytab Server allows you to **explore in Tableau data which is generated
dynamically by a Jupyter Notebook**. You can thus create Tableau data
sources in a very flexible way using all the power of Python.
This is achieved by having Tableau access data through a **web
server created by Jupytab**.

Jupytab Server is built on **solid foundations**: Tableau's Web Data
Connector and the Jupyter Kernel Gateway.

Features:

* **Expose multiple pandas dataframes** to Tableau from a Jupyter
  notebook
* Access **several notebooks** from Tableau through a **single
  entry point** (web server)
* Manage your notebooks using a **web interface**
* **Secure access** to your data

The full documentation is available on the project's home page.
"""

# This call to setup() does all the work
setup(
    name="jupytab-server",
    version=VERSION,
    description="Connect Tableau to your Jupyter Notebook",
    long_description_content_type="text/markdown",
    long_description=README,
    url="https://github.com/CFMTech/Jupytab",
    author="Brian Tribondeau",
    author_email="brian.tribondeau@cfm.fr",
    license="MIT",
    keywords="jupytab jupyter notebook wdc tableau",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(exclude=["*.tests", "*.tests.*"]),
    python_requires='>=3.6',
    include_package_data=True,
    install_requires=["jupyter_kernel_gateway"],
    project_urls={
        "Bug Tracker": "https://github.com/CFMTech/Jupytab/issues",
        "Documentation": "https://github.com/CFMTech/Jupytab",
        "Source Code": "https://github.com/CFMTech/Jupytab"
    },
    entry_points={
        "console_scripts": [
            "jupytab=jupytab_server.jupytab:main",
        ]
    },
)



# Copyright (c) 2019 Capital Fund Management
# SPDX-License-Identifier: MIT

import logging
import os
import socket
import subprocess
import sys

from . import log_pipe


class KernelCallback:
    def on_kernel_start(self, kernel_executor):
        pass

    def on_kernel_stop(self, kernel_executor):
        pass


class KernelExecutor:
    def __init__(self, name, file_path, description=None, kernel_callback=None, cwd=None):
        self.__notebook_name = name
        self.__notebook_file = file_path
        self.__notebook_description = description
        self.__kernel_callback = kernel_callback
        self.__kernel_process = None
        self.__host = '127.0.0.1'
        self.__port = None
        self.__cwd = cwd

    @staticmethod
    def get_free_tcp_port():
        tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        tcp.bind(('', 0))
        host, port = tcp.getsockname()
        tcp.close()

        return port

    def start(self):
        my_env = dict(os.environ)
        my_env['PATH'] = os.path.dirname(sys.executable) + os.pathsep + os.environ.get('PATH')

        self.__port = KernelExecutor.get_free_tcp_port()

        print(f'Start notebook {self.notebook_file} on {self.host}:{self.port}')

        logpipe = log_pipe.LogPipe(logging.INFO)

        command_line = [
            u"jupyter",
            u"kernelgateway",
            u"--KernelGatewayApp.api='kernel_gateway.notebook_http'",
            u"--KernelGatewayApp.seed_uri={filename}".format(filename=self.notebook_file),
            u"--KernelGatewayApp.ip='{host}'".format(host=self.host),
            u"--KernelGatewayApp.port={port}".format(port=self.port)
        ]

        self.__kernel_process = subprocess.Popen(command_line,
                                                 env=my_env,
                                                 stdout=logpipe,
                                                 stderr=logpipe,
                                                 cwd=self.cwd)

        if self.__kernel_callback:
            self.__kernel_callback.on_kernel_start(self)

        return f'{self.host}:{self.port}'

    def stop(self):
        print(f'Stop notebook {self.notebook_file} on {self.host}:{self.port}')

        self.__kernel_process.kill()
        self.__kernel_process = None

        if self.__kernel_callback:
            self.__kernel_callback.on_kernel_stop(self)

    def restart(self):
        print(f'Restart notebook {self.notebook_file} on {self.host}:{self.port}')

        self.stop()
        self.start()

    @property
    def notebook_file(self):
        return self.__notebook_file

    @property
    def kernel_process(self):
        return self.__kernel_process

    @property
    def name(self):
        return self.__notebook_name

    @property
    def host(self):
        return self.__host

    @property
    def port(self):
        return self.__port

    @property
    def description(self):
        return self.__notebook_description

    @property
    def cwd(self):
        return self.__cwd

    @property
    def kernel_status(self):
        if not self.__kernel_process:
            return "UNKNOWN"
        elif not self.__kernel_process.returncode:
            return "RUNNING"
        else:
            return f"STOPPED ({self.__kernel_process.returncode})"
